from pathlib import Path
from git import InvalidGitRepositoryError, Repo, GitCommandError
import sys
import stat
import inspect
import os
from typing import List
import shutil
import yaml

from .errors import PipelineExistsError, PipelineNotFoundError, InvalidPipelineRepositoryError
from .cli.config import SnkConfig
from .cli.pipeline import Pipeline


class Nest:
    """Class for managing Snk pipelines"""

    def __init__(self, snk_home: Path = None, bin_dir: Path = None) -> None:
        self.python_interpreter_path = Path(sys.executable) # needs to be the same python that has snk
        
        if not snk_home:
            home_path = self.python_interpreter_path.parent.parent
            if not os.access(home_path, os.W_OK):
                user_home_path = Path('~').expanduser()
                snk_home = user_home_path / ".local" / 'snk'
            else:
                snk_home = home_path / 'snk'

        if not bin_dir:
            bin_dir = self.python_interpreter_path.parent
            if not os.access(bin_dir, os.W_OK):
                user_home_path = Path('~').expanduser()
                bin_dir = user_home_path / ".local" / 'bin'

        self.snk_home = Path(snk_home).absolute()
        self.pipelines_dir = self.snk_home / "pipelines"
        self.bin_dir = Path(bin_dir).absolute()
        
        # Create dirs
        self.snk_home.mkdir(parents=True, exist_ok=True)
        self.pipelines_dir.mkdir(parents=True, exist_ok=True)
        self.bin_dir.mkdir(parents=True, exist_ok=True)

    def _check_repo_url_format(self, repo: str):
        if not repo.startswith('http'):
            raise InvalidPipelineRepositoryError('Repo url must start with http')
        if not repo.endswith('.git'):
            raise InvalidPipelineRepositoryError('Repo url must end in .git')

    def install(self, pipeline: str, editable = False, name = None, tag = None, config: Path = None, force = False, resources=[]) -> Pipeline:
        """
        Install a Snakemake pipeline as a CLI. 
        They must be standards compliant*, public, Snakemake workflows.
        https://snakemake.github.io/snakemake-workflow-catalog/
        """
        try:
            self._check_repo_url_format(pipeline)
            if not name:
                name = self._get_name_from_git_url(pipeline)
            if name in os.listdir(self.pipelines_dir):
                raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.pipelines_dir}")
            if name in os.listdir(self.bin_dir):
                raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.bin_dir}")
            repo = self.download(pipeline, name, tag_name=tag)
        except InvalidPipelineRepositoryError:
            pipeline = Path(pipeline)
            if pipeline.suffix == '.snk':
                pipeline = pipeline.parent
            if not name:
                name = pipeline.name
            repo = self.local(pipeline, name, editable)
        try:
            pipeline = Pipeline(path=Path(repo.git_dir).parent)
            pipeline_executable = self.create_package(pipeline.path)
            self.link_pipeline_executable_to_bin(pipeline_executable)
            if config:
                self.copy_nonstandard_config(pipeline.path, config)
            if resources:
                self.additional_resources(pipeline.path, resources)
            self._confirm_installation(name)
        except Exception as e:
            # remove any half completed steps 
            to_remove = self.get_paths_to_delete(name)
            self.delete_paths(to_remove)
            raise e
        return pipeline
    
    def additional_resources(self, pipeline_dir: Path, resources: List[Path]):
        """Modify the .snk file so that resources will be copied at runtime."""
        # validate_resources(resources)
        snk_config = SnkConfig.from_path(pipeline_dir / '.snk')
        snk_config.resources += [r for r in resources if r not in snk_config.resources]
        snk_config.to_yaml(pipeline_dir / '.snk')

    def copy_nonstandard_config(self, pipeline_dir: Path, config_path: Path):
        config_dir = pipeline_dir / 'config'
        config_dir.mkdir()
        shutil.copyfile(pipeline_dir / config_path, config_dir / 'config.yaml')

    def get_paths_to_delete(self, pipeline_name: str) -> List[Path]:
        to_delete = []
        
        # remove repo 
        pipeline_dir = self.pipelines_dir / pipeline_name
        if pipeline_dir.exists() and pipeline_dir.is_dir():
            to_delete.append(pipeline_dir)
        else:
            raise PipelineNotFoundError(f'Could not find pipeline: {pipeline_name}')

        # remove link
        pipeline_bin_executable = self.bin_dir / pipeline_name
        if pipeline_bin_executable.exists() and pipeline_bin_executable.is_symlink():
            if str(pipeline_dir) in str(os.readlink(pipeline_bin_executable)):
                to_delete.append(pipeline_bin_executable)
        
        return to_delete

    def delete_paths(self, files: List[Path]):
        # check that the files are in self.pipelines_dir
        # i.e. if it is a symlink read the link and check
        for path in files:
            if path.is_dir():
                assert str(self.snk_home) in str(path), "Cannot delete folders outside of SNK_HOME"
                shutil.rmtree(path)
            elif path.is_symlink():
                assert str(self.snk_home) in str(os.readlink(path)), "Cannot delete files outside of SNK_HOME"
                path.unlink()
            else:
                raise TypeError("Invalid file type")

    def uninstall(self, name: str, force: bool = False) -> bool:
        to_remove = self.get_paths_to_delete(name)
        if force:
            proceed = True
        else:
            print(f"Uninstalling {name}")
            print("  Would remove:")
            for p in to_remove:
                print(f"    {p}{'/*' if p.is_dir() else ''}")
            ans = input("Proceed (Y/n)? ")
            proceed = ans.lower() in ['y', 'yes']
        if not proceed:
            return False
        self.delete_paths(to_remove)
        return True 

    def _confirm_installation(self, name: str):
        pipeline_dir = self.pipelines_dir / name
        assert pipeline_dir.exists()
        pipelines = [p.name.split('.')[0] for p in self.bin_dir.glob('*')]
        assert name in pipelines

    def _get_name_from_git_url(self, git_url: str):
        return git_url.split('/')[-1].split('.')[0]

    @property
    def pipelines(self):
        return [Pipeline(pipeline_dir.resolve()) for pipeline_dir in self.pipelines_dir.glob('*') if pipeline_dir.is_dir()]

    def download(self, repo_url: str, name: str, tag_name: str = None) -> Repo:
        """Pull the repo"""
        location  = self.pipelines_dir / name
        options = ['--depth 1', '--single-branch']
        if tag_name:
            options.append(f'--branch {tag_name}')
        try:
            repo = Repo.clone_from(repo_url, location, multi_options=options)
            repo.git.checkout(tag_name)
        except GitCommandError as e:
            if "destination path" in e.stderr:
                raise PipelineExistsError(f"Pipeline '{name}' already exists in {self.pipelines_dir}")
            elif f"Remote branch {tag_name}" in e.stderr:
                raise PipelineNotFoundError(f"Pipeline tag '{tag_name}' not found")
            elif "not found" in e.stderr:
                raise PipelineNotFoundError(f"Pipeline repository '{repo_url}' not found")
            raise e
        return repo
    
    def local(self, path: Path, name: str, editable=False):
        location  = self.pipelines_dir / name
        if editable:
            os.symlink(path, location)
        else:
            shutil.copytree(path, location)
        try:
            return Repo(location)
        except InvalidGitRepositoryError:
            return Repo.init(location, mkdir=False)
        


    def create_package(self, pipeline_dir: Path) -> Path:
        """Convert a SnakeMake pipeline repo into a snk CLI"""
        self.validate_SnakeMake_repo(pipeline_dir)
        
        template = inspect.cleandoc(f"""
            #!/bin/sh
            '''exec' "{self.python_interpreter_path}" "$0" "$@"
            ' '''
            # -*- coding: utf-8 -*-
            import re
            import sys
            from snk import create_cli
            if __name__ == "__main__":
                sys.argv[0] = re.sub(r'(-script\.pyw|\.exe)?$', '', sys.argv[0])
                sys.exit(create_cli("{pipeline_dir}"))
                
        """)

        pipeline_bin_dir = pipeline_dir / 'bin'
        pipeline_bin_dir.mkdir(exist_ok=True)

        name = pipeline_dir.name

        if sys.platform.startswith('win'):
            name += '.exe'
        
        pipeline_executable = pipeline_bin_dir / name

        with open(pipeline_executable, 'w') as f:
            f.write(template)

        pipeline_executable.chmod(pipeline_executable.stat().st_mode | stat.S_IEXEC)

        return pipeline_executable

    def link_pipeline_executable_to_bin(self, pipeline_executable_path: Path):
        name = pipeline_executable_path.name
        os.symlink(pipeline_executable_path.absolute(), self.bin_dir / name)
        return self.bin_dir / name

    def validate_SnakeMake_repo(self, repo: Repo):
        print("Skipping validation!")
        pass


from pathlib import Path
from snk.cli.utils import flatten
import snakemake

def test_flatten(example_config: Path):
    config = snakemake.load_configfile(example_config)
    flat_config = flatten(config)
    assert flat_config['diffexp:contrasts:A-vs-B'] == ['A', 'B']

def test_install(basic_runner):
    res = basic_runner(['--help'])
    assert 'Usage: snk-basic-pipeline' in res.stdout

def test_info(basic_runner):
    res = basic_runner(['info'])
    assert 'snk-basic-pipeline' in res.stdout, res.stderr
    assert 'version' in res.stdout
    assert 'pipeline_dir_path' in res.stdout

# def test_run(basic_runner):
#     res = basic_runner(['info'])
#     assert 'snk-basic-pipeline' in res.stdout, res.stderr
#     assert 'version' in res.stdout
#     assert 'pipeline_dir_path' in res.stdout


from pathlib import Path
import sys
from git import Repo, GitCommandError

class Pipeline:
    """
    Represents a pipeline.
    Attributes:
      path (Path): The path to the pipeline.
      repo (Repo): The git repository of the pipeline.
      name (str): The name of the pipeline.
    """
    def __init__(self, path:Path) -> None:
        """
        Initializes a Pipeline object.
        Args:
            path (Path): The path to the pipeline.
        Returns:
            None
        Notes:
            Initializes the `repo` and `name` attributes.
        """
        self.path = path
        self.repo = Repo(path)
        self.name = self.path.name
    
    @property
    def version(self):
        """
        Gets the version of the pipeline.
        Returns:
            str: The version of the pipeline, or None if no version is found.
        """
        try:
            # TODO: default to commit
            version = self.repo.git.describe(['--tags','--exact-match']) 
        except GitCommandError:
            version = None
        return version

    @property
    def executable(self):
        """
        Gets the executable of the pipeline.
        Returns:
            Path: The path to the pipeline executable.
        """
        pipeline_bin_dir = self.path / 'bin'
        name = self.name
        if sys.platform.startswith('win'):
            name += '.exe'
        return pipeline_bin_dir / name
        



from pathlib import Path
from typing import List
import snakemake
from dataclasses import dataclass, field

import yaml

@dataclass
class SnkConfig:
    """
    SNK config holds dynamic cli config and option annotations.
    """
    resources: List[Path] = field(default_factory=list)
    annotations: dict = field(default_factory=dict)

    @classmethod
    def from_path(cls, snk_config_path: Path):
        """
        Load and validate SNK config from .snk file.
        """
        if snk_config_path.exists():
            snk_config_dict = snakemake.load_configfile(snk_config_path)
            snk_config = cls(**snk_config_dict)
            snk_config.resources = [snk_config_path.parent / resource for resource in snk_config.resources]
            snk_config.validate_resources(snk_config.resources)
            return snk_config
        return cls()
    
    def validate_resources(self, resources):
        """
    Validate resources.
    Args:
      resources (List[Path]): List of resources to validate.
    Raises:
      FileNotFoundError: If a resource is not found.
    Notes:
      This function does not modify the resources list.
    """
        for resource in resources:
            assert resource.exists(), FileNotFoundError(f"Could not find resource: {resource}")

    def add_resources(self, resources: List[Path], pipeline_dir_path: Path = None):
        """
        Add and validate resources. 
        It takes a list of resources, and if the resource is not absolute, it appends the
        pipeline_dir_path to the resource
        
        :param resources: List[Path]
        :param pipeline_dir_path: The path to the directory containing the pipeline
        """
        processed = []
        for resource in resources:
            if pipeline_dir_path and not resource.is_absolute():
                resource = pipeline_dir_path / resource
            processed.append(resource)
        self.validate_resources(processed)
        self.resources.extend(processed)

    def to_yaml(self, path: Path) -> None:
        """
        Write SNK config to YAML file.
        Args:
            path (Path): Path to write the YAML file to.
        Returns:
            None
        Side Effects:
            Writes the SNK config to the specified path.
        """
        with open(path, 'w') as f:
            yaml.dump(vars(self), f)
                             
    

def get_config_from_pipeline_dir(pipeline_dir_path: Path):
    """Search possible config locations"""
    for path in [Path('config') / 'config.yaml', Path('config') / 'config.yml', 'config.yaml', 'config.yml']:
        if (pipeline_dir_path / path).exists():
            return pipeline_dir_path / path 
    return None



def load_pipeline_snakemake_config(pipeline_dir_path: Path):
    """
    Load snakemake config.
    """
    pipeline_config_path = get_config_from_pipeline_dir(pipeline_dir_path)
    if not pipeline_config_path.exists():
        return []
    return snakemake.load_configfile(pipeline_config_path)



from typing import List, Callable
from inspect import signature, Parameter
from makefun import wraps
from pathlib import Path
import typer
import sys
import collections  # MutableMapping import hack
if sys.version_info.major == 3 and sys.version_info.minor >= 10:
    from collections.abc import MutableMapping
else:
    from collections import MutableMapping


types = {
    'int': int,
    'integer': int,
    'str': str,
    'string': str,
    'path': Path,
    'bool': bool,
    'boolean': bool,
    'list': List[str],
    'list[path]': List[Path],
    'list[int]': List[int],
}

def create_cli_parameter(option):
    """
    Creates a parameter for a CLI option.
    Args:
      option (dict): A dictionary containing the option's name, type, required status, default value, and help message.
    Returns:
      Parameter: A parameter object for the CLI option.
    Examples:
      >>> option = {'name': 'foo', 'type': 'int', 'required': True, 'default': 0, 'help': 'A number'}
      >>> create_cli_parameter(option)
      Parameter('foo', kind=Parameter.POSITIONAL_OR_KEYWORD, default=typer.Option(..., help='[CONFIG] A number'), annotation=int)
    """
    return Parameter(
        option['name'], 
        kind=Parameter.POSITIONAL_OR_KEYWORD, 
        default=typer.Option(... if option['required'] else option['default'], help=f"[CONFIG] {option['help']}"), 
        annotation=types.get(option['type'].lower(), str))

def add_dynamic_options(options: List[dict]):
    """
    Decorator to add dynamic options to a function.
    Args:
      options (List[dict]): A list of dictionaries containing the option's name, type, required status, default value, and help message.
    Returns:
      Callable: A decorated function with the dynamic options added.
    Examples:
      >>> @add_dynamic_options([{'name': 'foo', 'type': 'int', 'required': True, 'default': 0, 'help': 'A number'}])
      ... def my_func(ctx):
      ...     pass
      >>> my_func
      <function my_func at 0x7f8f9f9f9f90>
    """
    def inner(func: Callable):
        """
    Wraps a function with dynamic options.
    Args:
      func (Callable): The function to wrap.
    Returns:
      Callable: A wrapped function with the dynamic options added.
    Notes:
      This function is used as an inner function in the `add_dynamic_options` decorator.
    """
        func_sig = signature(func)
        params = list(func_sig.parameters.values())
        for op in options[::-1]:
            params.insert(1, create_cli_parameter(op))
        new_sig = func_sig.replace(parameters=params)
        @wraps(func, new_sig=new_sig)
        def func_wrapper(*args, **kwargs):
            """
            Wraps a function with dynamic options.
            Args:
                *args: Variable length argument list.
                **kwargs: Arbitrary keyword arguments.
            Returns:
                Callable: A wrapped function with the dynamic options added.
            Notes:
                This function is used as an inner function in the `add_dynamic_options` decorator.
            """
            if kwargs['configfile']:
                # need to check if kwargs in options have changed
                # parse the new configfile and update the defautls
                raise NotImplementedError
            for op in options:
                kwargs['ctx'].args.extend([f"--{op['name']}", kwargs[op['name']]])
            kwargs= {k:v for k,v in kwargs.items() if k in func_sig.parameters.keys()}
            return func(*args, **kwargs)
        return func_wrapper
    return inner


def flatten(d, parent_key='', sep=':'):
    """
    Flattens a nested dictionary.
    Args:
      d (dict): The dictionary to flatten.
      parent_key (str, optional): The parent key of the dictionary. Defaults to ''.
      sep (str, optional): The separator to use between keys. Defaults to ':'.
    Returns:
      dict: A flattened dictionary.
    Examples:
      >>> d = {'a': {'b': 1, 'c': 2}, 'd': 3}
      >>> flatten(d)
      {'a:b': 1, 'a:c': 2, 'd': 3}
    """
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, MutableMapping):
            items.extend(flatten(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


import typer
from pathlib import Path
from typing import Optional, List, Callable
from datetime import datetime
import subprocess
import shutil
import os
from contextlib import contextmanager

import snakemake
from rich.console import Console
from rich.syntax import Syntax
from art import text2art



from .config import SnkConfig, get_config_from_pipeline_dir, load_pipeline_snakemake_config
from .utils import add_dynamic_options, flatten
from .pipeline import Pipeline


def convert_key_to_samkemake_format(key, value):
    """
    Covert key to a format that can be passed over the cli to snakemake
    """
    resultDict = dict()
    parts = key.split(":")
    d = resultDict
    for part in parts[:-1]:
        if part not in d:
            d[part] = dict()
        d = d[part]
    d[parts[-1]] = value
    return resultDict

def serialise(d):
    """
    Serialises a data structure into a string.
    Args:
      d (any): The data structure to serialise.
    Returns:
      any: The serialised data structure.
    Examples:
      >>> serialise({'a': 1, 'b': 2})
      {'a': '1', 'b': '2'}
    """
    if isinstance(d, Path) or isinstance(d, datetime):
        return str(d)

    if isinstance(d, list):
        return [serialise(x) for x in d]

    if isinstance(d, dict):
        for k, v in d.items():
            d.update({k: serialise(v)})

    # return anything else, like a string or number
    return d

def parse_config_args(args: List[str], options):
    """
    Parses a list of arguments and a list of options.
    Args:
      args (List[str]): A list of arguments.
      options (List[dict]): A list of options.
    Returns:
      (List[str], List[dict]): A tuple of parsed arguments and config.
    Examples:
      >>> parse_config_args(['-name', 'John', '-age', '20'], [{'name': 'name', 'default': '', 'help': '', 'type': 'str', 'required': True}, {'name': 'age', 'default': '', 'help': '', 'type': 'int', 'required': True}])
      (['John', '20'], [{'name': 'name', 'John'}, {'age': 20}])
    """
    names = [op['name'] for op in options]
    config = []
    parsed = []
    flag=None
    for arg in args:
        if flag:
            name = flag.lstrip('-')
            op = next(op for op in options if op['name'] == name)
            if op['default'] == serialise(arg):
                # skip args that don't change
                flag=None
                continue
            if ":" in op['original_key']:
                samkemake_format_config = convert_key_to_samkemake_format(op['original_key'], arg)
                name = list(samkemake_format_config.keys())[0]
                arg = samkemake_format_config[name]
            # config.append(f'{name}={serialise(arg)}')
            config.append({name: serialise(arg)})
            flag=None
            continue
        if arg.startswith('-') and arg.lstrip('-') in names:
            flag = arg
            continue
        parsed.append(arg)
    return parsed, config


def build_dynamic_cli_options(snakemake_config, snk_config: SnkConfig):
    """
    Builds a list of options from a snakemake config and a snk config.
    Args:
      snakemake_config (dict): A snakemake config.
      snk_config (SnkConfig): A snk config.
    Returns:
      List[dict]: A list of options.
    Examples:
      >>> build_dynamic_cli_options({'name': 'John', 'age': 20}, {'annotations': {'name:name': 'name', 'name:help': '', 'name:type': 'str', 'name:required': True, 'age:name': 'age', 'age:help': '', 'age:type': 'int', 'age:required': True}})
      [{'name': 'name', 'original_key': 'name', 'default': 'John', 'help': '', 'type': 'str', 'required': True}, {'name': 'age', 'original_key': 'age', 'default': 20, 'help': '', 'type': 'int', 'required': True}]
    """
    flat_config = flatten(snakemake_config)
    options = []
    flat_snk_annotations = flatten(snk_config.annotations)
    for op in flat_config:
        name = flat_snk_annotations.get(f"{op}:name", op.replace(':', '_'))
        help = flat_snk_annotations.get(f"{op}:help", "")
        # TODO be smarter here 
        # look up the List type e.g. if type == list then check the frist index type 
        # also can probably just pass the type around instead of the string?
        param_type = flat_snk_annotations.get(f"{op}:type", f"{type(flat_config[op]).__name__}")  # TODO refactor 
        required = flat_snk_annotations.get(f"{op}:required", False)
        options.append(
            {
                'name':name.replace('-', '_'),
                'original_key': op,
                'default': flat_config[op],
                'help': help,
                'type': param_type,
                'required': required
            }
        )
    # TODO: find annotations missing from config and add them to options
    return options


class CLI:
    """
    Constructor for the CLI class.
    Args:
      pipeline_dir_path (Path): Path to the pipeline directory.
    Side Effects:
      Initializes the CLI class.
    Examples:
      >>> CLI(Path('/path/to/pipeline'))
    """
    def __init__(self, pipeline_dir_path: Path) -> None:
        self.pipeline = Pipeline(path=pipeline_dir_path)
        self.app = typer.Typer()
        self.snakemake_config = load_pipeline_snakemake_config(pipeline_dir_path)
        self.snk_config: SnkConfig = SnkConfig.from_path(pipeline_dir_path / '.snk')
        self.options = build_dynamic_cli_options(self.snakemake_config, self.snk_config)
        self.snakefile = self._find_snakefile()
        self.conda_prefix_dir = pipeline_dir_path / '.conda'
        self.name = self.pipeline.name
        def _print_pipline_version(ctx: typer.Context, value: bool):
            if value:
                typer.echo(self.pipeline.version)
                raise typer.Exit()

        def _print_pipline_path(ctx: typer.Context, value: bool):
            if value:
                typer.echo(self.pipeline.path)
                raise typer.Exit()

        def callback(
            ctx: typer.Context, 
            version: Optional[bool] = typer.Option(None, '-v', '--version', help="Show the pipeline version.", is_eager=True, callback=_print_pipline_version, show_default=False),
            path: Optional[bool] = typer.Option(None, '-p', '--path', help="Show the pipeline path.", is_eager=True, callback=_print_pipline_path, show_default=False)
        ):
            if ctx.invoked_subcommand is None:
                typer.echo(f'{ctx.get_help()}')
        # dynamically create the logo
        callback.__doc__ = f"{self.create_logo()}"

        # registration 
        self.register_callback(callback, invoke_without_command=True, context_settings={"help_option_names": ["-h", "--help"]})
        self.register_command(self.info, help="Display information about current pipeline install.")
        self.register_command(self.config, help="Access the pipeline configuration.")
        self.register_command(self.env, help="Access the pipeline conda environments.")
        self.register_command(self.script, help="Access the pipeline scripts.")
        self.register_command(
            add_dynamic_options(self.options)(self.run), 
            help="Run the pipeline. All unrecognized arguments are parsed onto Snakemake to be used by the pipeline.", 
            context_settings={
                "allow_extra_args": True, 
                "ignore_unknown_options": True, 
                "help_option_names": ["-h", "--help"]
            }
        )

    def __call__(self):
        """
    Invoke the CLI.
    Side Effects:
      Invokes the CLI.
    Examples:
      >>> CLI(Path('/path/to/pipeline'))()
    """
        self.app()

    def register_command(self, command: Callable, **command_kwargs) -> None:
        """
    Register a command to the CLI.
    Args:
      command (Callable): The command to register.
    Side Effects:
      Registers the command to the CLI.
    Examples:
      >>> CLI.register_command(my_command)
    """
        self.app.command(**command_kwargs)(command)

    def register_callback(self, command: Callable, **command_kwargs) -> None:
        """
    Register a callback to the CLI.
    Args:
      command (Callable): The callback to register.
    Side Effects:
      Registers the callback to the CLI.
    Examples:
      >>> CLI.register_callback(my_callback)
    """
        self.app.callback(**command_kwargs)(command)

    def create_logo(self, font="small"):
        """
    Create a logo for the CLI.
    Args:
      font (str): The font to use for the logo.
    Returns:
      str: The logo.
    Examples:
      >>> CLI.create_logo()
    """
        logo = text2art(self.name, font=font)        
        doc  = f"""\b{logo}\bA Snakemake pipeline CLI generated with snk"""
        return doc

    def _print_snakemake_help(value: bool):
        """
    Print the snakemake help and exit.
    Args:
      value (bool): If True, print the snakemake help and exit.
    Side Effects:
      Prints the snakemake help and exits.
    Examples:
      >>> CLI._print_snakemake_help(True)
    """
        if value:
            snakemake.main("-h")
    
    def _find_snakefile(self):
            """
    Search possible snakefile locations.
    Returns:
      Path: The path to the snakefile.
    Examples:
      >>> CLI._find_snakefile()
    """
            for path in snakemake.SNAKEFILE_CHOICES:
                if (self.pipeline.path / path).exists():
                    return self.pipeline.path / path 
            raise FileNotFoundError("Snakefile not found!")
    
    @contextmanager
    def copy_resources(self, resources: List[Path], cleanup: bool):
        """
    Copy resources to the current working directory.
    Args:
      resources (List[Path]): A list of paths to the resources to copy.
      cleanup (bool): If True, the resources will be removed after the function exits.
    Side Effects:
      Copies the resources to the current working directory.
    Returns:
      Generator: A generator object.
    Examples:
      >>> with CLI.copy_resources(resources, cleanup=True):
      ...     # do something
    """
        copied_resources = []

        def copy_resource(src, dst):
            if src.is_dir():
                shutil.copytree(src, dst)
            else:
                shutil.copy(src, dst)

        def remove_resource(resource):
            if resource.is_dir():
                shutil.rmtree(resource)
            else:
                os.remove(resource)

        try:
            for resource in resources:
                abs_path = self.pipeline.path / resource
                destination = Path('.') / resource.name
                if not destination.exists(): 
                    # make sure you don't delete files that are already there...
                    copy_resource(abs_path, destination)
                    copied_resources.append(destination)
                else:
                    typer.secho(
                        f"Resource {resource.name} already exists... Skipping!", 
                        fg=typer.colors.YELLOW
                    )

            yield
        finally:
            if not cleanup:
                return 
            for copied_resource in copied_resources:
                if copied_resource.exists():
                    remove_resource(copied_resource)

    def run(
            self,
            ctx: typer.Context,
            target: str = typer.Argument(None, help="File to generate. If None will run the pipeline 'all' rule."),
            configfile: Path = typer.Option(None, help="Path to snakemake config file. Overrides existing config and defaults.", exists=True, dir_okay=False),
            resource: List[Path] = typer.Option([], help="Additional resources to copy to workdir at run time."),
            cleanup_resources: Optional[bool] = typer.Option(True, help="Delete resources once the pipeline sucessfully completes."),
            cleanup_snakemake: Optional[bool] = typer.Option(True, help="Delete .snakemake folder once the pipeline sucessfully completes."),
            cores:  int = typer.Option(None, help="Set the number of cores to use. If None will use all cores."),
            verbose: Optional[bool] = typer.Option(False, "--verbose", "-v", help="Run pipeline in verbose mode.",),
            help_snakemake: Optional[bool] = typer.Option(
                False, "--help-snakemake", "-hs", help="Print the snakemake help and exit.", is_eager=True, callback=_print_snakemake_help, show_default=False
            ),
        ):
        """
    Run the pipeline.
    Args:
      target (str): File to generate. If None will run the pipeline 'all' rule.
      configfile (Path): Path to snakemake config file. Overrides existing config and defaults.
      resource (List[Path]): Additional resources to copy to workdir at run time.
      cleanup_resources (bool): Delete resources once the pipeline sucessfully completes.
      cleanup_snakemake (bool): Delete .snakemake folder once the pipeline sucessfully completes.
      cores (int): Set the number of cores to use. If None will use all cores.
      verbose (bool): Run pipeline in verbose mode.
      help_snakemake (bool): Print the snakemake help and exit.
    Side Effects:
      Runs the pipeline.
    Examples:
      >>> CLI.run(target='my_target', configfile=Path('/path/to/config.yaml'), resource=[Path('/path/to/resource')], verbose=True)
    """
        args = []
        if not cores:
            cores = 'all'
        args.extend([
            "--use-conda",
            f"--conda-prefix={self.conda_prefix_dir}",
            f"--cores={cores}",
        ])
        if not self.snakefile.exists():
            raise ValueError('Could not find Snakefile') # this should occur at install
        else:
            args.append(f"--snakefile={self.snakefile}")
        
        if not configfile:
            configfile = get_config_from_pipeline_dir(self.pipeline.path)
        if configfile:
            args.append(f"--configfile={configfile}")

        
        # Set up conda frontend
        mamba_found = True
        try:
            subprocess.run(["mamba", "--version"], capture_output=True, check=True)
        except (subprocess.CalledProcessError, FileNotFoundError):
            typer.secho("Mamba not found! Install for speed up.", fg=typer.colors.YELLOW)
            mamba_found = False
        if not mamba_found:
            args.append("--conda-frontend=conda")
        
        typer.echo(self.create_logo())
        typer.echo()

        if verbose:
            args.insert(0, "--verbose")

        if target:
            args.append(target)
        targets_and_or_snakemake, config_dict_list = parse_config_args(ctx.args, options=self.options)

        args.extend(targets_and_or_snakemake)

        args.extend(["--config", *[f"{list(c.keys())[0]}={list(c.values())[0]}" for c in config_dict_list]])
        if verbose:
            typer.secho(f"snakemake {' '.join(args)}\n", fg=typer.colors.MAGENTA)
        
        self.snk_config.add_resources(resource, self.pipeline.path)
        with self.copy_resources(self.snk_config.resources, cleanup=cleanup_resources):
                snakemake.main(args)
        if cleanup_snakemake:
            shutil.rmtree(".snakemake")

    def info(self):
        """
    Display information about current pipeline install.
    Returns:
      str: A JSON string containing information about the current pipeline install.
    Examples:
      >>> CLI.info()
    """
        import json
        info_dict = {}
        info_dict['name'] = self.pipeline.path.name
        info_dict['version'] = self.pipeline.version
        info_dict['pipeline_dir_path'] = str(self.pipeline.path)
        typer.echo(json.dumps(info_dict, indent=2))

    def config(self):
        """
    Access the pipeline configuration.
    Side Effects:
      Prints the pipeline configuration.
    Examples:
      >>> CLI.config()
    """
        config_path = get_config_from_pipeline_dir(self.pipeline.path)
        if not config_path:
            typer.secho("Could not find config...", fg='red')
            raise typer.Exit(1)
        with open(config_path) as f:
            code = f.read()
            syntax = Syntax(code, 'yaml')
            console = Console()
            console.print(syntax)
    
    def env(
        name: Optional[str] = typer.Argument(None)
    ):
        """
    Access the pipeline conda environments.
    Args:
      name (str): The name of the environment.
    Examples:
      >>> CLI.env(name='my_env')
    """
        raise NotImplementedError

    def script(
        name: Optional[str] = typer.Argument(None)
    ):
        """
    Access the pipeline scripts.
    Args:
      name (str): The name of the script.
    Examples:
      >>> CLI.script(name='my_script')
    """
        raise NotImplementedError


from pathlib import Path
from .main import CLI

def create_cli(p):
    """
    This is the interface to create the dynamic CLI. 
    Changing this function name will break backwards compatibility. 
    """
    pipeline_dir_path = Path(p)
    cli = CLI(pipeline_dir_path)
    cli()

